Cable = require("libraries/classic"):extend()

local radius = 7

function Cable:new(out_node, in_node, schema)
    self.out_node = out_node
    self.in_node = in_node
    self.schema = schema
    self.head_circle = Circle(self.out_node:get_output_x(), self.out_node:get_output_y(), radius)
    self.tail_circle = Circle(love.mouse.getX(), love.mouse.getY(), radius)
    self.mid_circle = Circle(0, 0, 10)
    self.dotted_line = DottedLine(3, 5)

    if self:is_complete() then
        self.tail_circle:move_to(self.in_node:get_input_x(), self.in_node:get_input_y())
    end

    self:update_mid_circle_position()
end

function Cable:update_mid_circle_position()
    self.mid_circle.x = 0.5 * (self.head_circle.x + self.tail_circle.x)
    self.mid_circle.y = 0.5 * (self.head_circle.y + self.tail_circle.y)
end

function Cable:is_complete()
    return self.in_node ~= nil
end

function Cable:set_in_node(in_node)
    self.in_node = in_node
end

function Cable:update()
    self.dotted_line.phase_pixels = love.timer.getTime() * 10

    self.head_circle:move_to(self.out_node:get_output_x(), self.out_node:get_output_y())

    if self.is_tail_dragged then
        self:set_in_node(nil)
        self.tail_circle:move_to(love.mouse.getPosition())
    else
        self.tail_circle:move_to(self.in_node:get_input_x(), self.in_node:get_input_y())
    end

    self:update_mid_circle_position()

    if input:pressed("pick") then
        if self.tail_circle:is_point_inside(love.mouse.getPosition()) then
            self.is_tail_dragged = true
        elseif self:is_complete() and self.mid_circle:is_point_inside(love.mouse.getPosition()) then
            self.schema:delete_cable(self)
        end
    end

    if input:released("pick") then
        if self.is_tail_dragged then
            self.schema:rewire_cable(self)
        end

        self.is_head_dragged = false
        self.is_tail_dragged = false
    end
end

function Cable:draw_dotted_line(x1, y1, x2, y2, dot_diameter, spacing, phase_pixels)
    local length = gamekit.vector.distance(x1, y1, x2, y2)
    local pattern_length = dot_diameter + spacing
    local advancement = math.fmod(phase_pixels or 0, pattern_length)
    local norm_x, norm_y = gamekit.vector.normalized(x2 - x1, y2 - y1)

    love.graphics.setLineWidth(1)
    while advancement <= length do
        local dot_x, dot_y = x1 + norm_x * advancement, y1 + norm_y * advancement

        love.graphics.circle("fill", dot_x, dot_y, 0.5 * dot_diameter)
        love.graphics.circle("line", dot_x, dot_y, 0.5 * dot_diameter, 10)

        advancement = advancement + pattern_length
    end

    return advancement
end

function Cable:draw_straight_line()
    local x1, y1 = self.head_circle.x, self.head_circle.y
    local x2, y2 = self.tail_circle.x, self.tail_circle.y
    --love.graphics.line(x1, y1, x2, y2)
    self:draw_dotted_line(x1, y1, x2, y2, 4, 8, love.timer.getTime() * 20)
end

function Cable:draw_bent_line()
    local x1, y1 = self.head_circle.x, self.head_circle.y
    local x2, y2 = self.tail_circle.x, self.tail_circle.y
    local middle_x = 0.5 * (x1 + x2)
    local middle_y = 0.5 * (y1 + y2)

    if y2 > y1 then
        --[[if x1 == x2 then
            love.graphics.line(x1, y1, x1, y2)
        else
            love.graphics.line(x1, y1,
            x1, middle_y, x2, middle_y,
            x2, y2)
        end--]]
        if x1 == x2 then
            self.dotted_line:draw_polyline(x1, y1, x1, y2)
        else
            self.dotted_line:draw_polyline(x1, y1,
                x1, middle_y, x2, middle_y,
                x2, y2)
        end
        --self:draw_dotted_line(x1, middle_y, x2, middle_y, 4, 6, love.timer.getTime() * 20)
        --self:draw_dotted_line(x1, y1, x1, middle_y, 4, 6, love.timer.getTime() * 20)
        --self:draw_dotted_line(x2, middle_y, x2, y2, 4, 6, love.timer.getTime() * 20)
    else
        --[[if y1 == y2 then
            love.graphics.line(x1, y1, x2, y1)
        else
            love.graphics.line(x1, y1, 
            middle_x, y1,
            middle_x, y2,
            x2, y2)
        end--]]
        self.dotted_line:draw_polyline(x1, y1, 
            middle_x, y1,
            middle_x, y2,
            x2, y2)
        --self:draw_dotted_line(middle_x, y1, middle_x, y2, 4, 6, love.timer.getTime() * 20)
        --self:draw_dotted_line(x1, y1, middle_x, y1, 4, 6, love.timer.getTime() * 20)
        --self:draw_dotted_line(middle_x, y2, x2, y2, 4, 6, love.timer.getTime() * 20)
    end
end

function Cable:draw()
    love.graphics.setLineWidth(4)
    love.graphics.setColor(0.6,0.6,0.6)
    self:draw_bent_line()
    --self:draw_straight_line()

    love.graphics.setLineWidth(2)

    love.graphics.setColor(0.6,0.6,0.6)
    self.head_circle:draw_fill()
    self.head_circle:draw_line(20)
    
    love.graphics.setColor(1,1,1)
    self.tail_circle:draw_fill()
    love.graphics.setColor(0.6,0.6,0.6)
    self.tail_circle:draw_line(20)

    if self:is_complete() and self.mid_circle:is_point_inside(love.mouse.getPosition()) then
        love.graphics.setColor(0.6,0.6,0.6)
        self.mid_circle:draw_fill()
        self.mid_circle:draw_line(20)
        love.graphics.setColor(1,1,1)
        love.graphics.line(self.mid_circle.x - 5, self.mid_circle.y - 5, self.mid_circle.x + 5, self.mid_circle.y + 5)
        love.graphics.line(self.mid_circle.x - 5, self.mid_circle.y + 5, self.mid_circle.x + 5, self.mid_circle.y - 5)
    end
end