local animation = {}

local pow = math.pow
local sin = math.sin
local pi = math.pi

function animation.tween(start_value, end_value, t)
    return start_value + t * (end_value - start_value)
end

local AnimationTimeline = {}
AnimationTimeline.__index = AnimationTimeline

function animation.new_timeline()
    local timeline = {
        tracks = {},
        length = 0,
    }
    setmetatable(timeline, AnimationTimeline)
    return timeline
end

function AnimationTimeline:get_length() return self.length end

local function sort_keyframes(a, b)
    return a.time_offset < b.time_offset
end

local function find_redundant_keyframe(timeline_track, time_offset)
    for i = 1, #timeline_track do
        if timeline_track[i].time_offset == time_offset then
            return i
        end
    end
end

function AnimationTimeline:set_track_owner_object(track_name, owner_object, destination_key)
    if self.tracks[track_name] == nil then
        self.tracks[track_name] = {
            owner_object = owner_object,
            destination_key = destination_key or track_name,
        }
    else
        local track = self.tracks[track_name]
        track.owner_object = owner_object
        track.destination_key = destination_key or track_name
    end
end

function AnimationTimeline:set_track_keyframe(track_name, time_offset, value, easing)
    if self.tracks[track_name] == nil then
        self.tracks[track_name] = {}
    else
        local redundant_keyframe_index = find_redundant_keyframe(self.tracks[track_name])
        if redundant_keyframe_index then
            table.remove(self.tracks[track_name], redundant_keyframe_index)
        end
    end
    table.insert(self.tracks[track_name], { 
        time_offset = time_offset, 
        value = value, 
        easing = easing or animation.ease_linear,
    })
    self.length = math.max(self.length, time_offset)
    table.sort(self.tracks[track_name], sort_keyframes)
end

function AnimationTimeline:get_track_hot_keyframes_at(track_name, time_offset)
    local track = self.tracks[track_name]
    local left_keyframe, right_keyframe
    local keyframe_count = #track

    for i = 1, keyframe_count do
        local keyframe = track[i]
        if keyframe.time_offset >= time_offset then
            right_keyframe = keyframe
            left_keyframe = track[i-1]
            break
        elseif i == keyframe_count then
            left_keyframe = keyframe
        end
    end
    
    return left_keyframe, right_keyframe
end

function AnimationTimeline:get_track_value_at(track_name, time_offset)
    local track = self.tracks[track_name]
    local keyframe_count = #track

    if keyframe_count == 0 then
        return nil
    elseif keyframe_count == 1 then
        return track[1].value
    else
        local left_keyframe, right_keyframe = self:get_track_hot_keyframes_at(track_name, time_offset)
        if left_keyframe and right_keyframe then
            local t = (time_offset - left_keyframe.time_offset)/(right_keyframe.time_offset - left_keyframe.time_offset)
            return animation.tween(left_keyframe.value, right_keyframe.value, right_keyframe.easing(t))
        elseif right_keyframe then
            return right_keyframe.value
        elseif left_keyframe then
            return left_keyframe.value
        end 
    end
end

function AnimationTimeline:apply_track_value_to_owner(track_name, time_offset)
    local track = self.tracks[track_name]
    if track.owner_object then
        track.owner_object[track.destination_key] = self:get_track_value_at(track_name, time_offset)
    end
end

local AnimationPlayer = {}
AnimationPlayer.__index = AnimationPlayer

function animation.new_player()
    local player = {
        timeline = nil,
        time_offset = 0,
        playback_speed = 1,
        playing = false,
    }
    setmetatable(player, AnimationPlayer)
    return player
end

function AnimationPlayer:play_timeline(timeline)
    if self.timeline ~= timeline then
        self.time_offset = 0
    end
    self.timeline = timeline
    self.playback_speed = 1
    self.playing = timeline ~= nil
end

function AnimationPlayer:play_timeline_backwards(timeline)
    self:play_timeline()
    self.playback_speed = -1
end

function AnimationPlayer:pause()
    self.playing = false
end

function AnimationPlayer:stop()
    self.time_offset = 0
    self.playing = false
end

function AnimationPlayer:is_playing() return self.playing end

function AnimationPlayer:get_timeline_track_value(track_name)
    if self.timeline then
        return self.timeline:get_track_value_at(track_name, self.playback_speed == -1 and self.timeline:get_length() - self.time_offset or self.time_offset)
    end
end

function AnimationPlayer:update(delta)
    if self.playing then
        local transformed_offset
        self.time_offset = self.time_offset + delta
        transformed_offset = self.playback_speed == -1 and math.max(0, self.timeline:get_length() - self.time_offset) or math.min(self.time_offset, self.timeline:get_length())

        for track_name in pairs(self.timeline.tracks) do
            self.timeline:apply_track_value_to_owner(track_name, transformed_offset)
        end

        if self.time_offset >= self.timeline:get_length() then
            if self.loops then
                self.time_offset = 0
            else
                self:stop()
            end
        end
    end
end

-- credits to https://easings.net/
-- all methods take the argument 'x' between 0 and 1 and return a value between 0 and 1

function animation.ease_linear(x) return x end

function animation.ease_step_in(x) return math.ceil(x) end
function animation.ease_step_out(x) return math.floor(x) end

function animation.ease_quad_in(x) return x*x end
function animation.ease_quad_out(x) return 1-(1-x)*(1-x) end
function animation.ease_quad_both(x) return x < 0.5 and (2*x*x) or (1 - pow(-2*x+2, 2)*0.5) end

function animation.ease_cubic_out(x) return 1 - pow(1 - x, 3) end

function animation.ease_sin(x) return sin((x * pi) * 0.5) end

function animation.ease_back_in(x) local c1 = 1.70158 local c3 = c1 + 1 return c3 * x * x * x - c1 * x * x end
function animation.ease_back_out(x) local c1 = 1.70158 local c3 = c1 + 1 return 1 + c3 * pow(x - 1, 3) + c1 * pow(x - 1, 2) end
function animation.ease_back_both(x) local c1 = 1.70158 local c2 = c1 * 1.525 return x < 0.5 and ((pow(2 * x, 2) * ((c2 + 1) * 2 * x - c2)) * 0.5) or ((pow(2 * x - 2, 2) * ((c2 + 1) * (x * 2 - 2) + c2) + 2) * 0.5) end

function animation.ease_elastic_in(x) local c4 = (2 * pi) / 3 return x == 0 and 0 or (x == 1 and 1 or (-pow(2, 10 * x - 10) * sin((x * 10 - 10.75) * c4))) end
function animation.ease_elastic_out(x) local c4 = (2 * pi) / 3 return x == 0 and 0 or (x == 1 and 1 or (pow(2, -10 * x) * sin((x * 10 - 0.75) * c4) + 1)) end
function animation.ease_elastic_both(x) local c5 = (2 * pi) / 4.5 return x == 0 and 0 or (x == 1 and 1 or (x < 0.5 and (-(pow(2, 20 * x - 10) * sin((20 * x - 11.125) * c5)) * 0.5) or ((pow(2, -20 * x + 10) * sin((20 * x - 11.125) * c5)) * 0.5 + 1))) end

function animation.ease_bounce(x)
    local n1 = 7.5625
    local d1 = 2.75

    if x < 1 / d1 then
        return n1 * x * x
    elseif x < 2 / d1 then
        return n1 * (x - (1.5 / d1)) * x + 0.75
    elseif x < 2.5 / d1 then
        return n1 * (x - (2.25 / d1)) * x + 0.9375
    else
        return n1 * (x - (2.625 / d1)) * x + 0.984375
    end
end



return animation