local check = require(PATH..'.check')
local rectangle = require(PATH..'.rectangle')
local viewport = require(PATH..'.viewport')
local camera = {}

local _x, _y = 0, 0
local _zoom = 1
local _movement_smoothing = 1
local _zooming_smoothing = 1
local _target_x, _target_y = 0, 0
local _target_zoom = 1
local _offset_x, _offset_y = 0, 0
local _pixel_perfect_enabled = false
local _transform = love.math.newTransform()

function camera.set_pixel_perfect_enabled(is_enabled)
    _pixel_perfect_enabled = is_enabled
end

function camera.get_pixel_perfect_enabled() 
    return _pixel_perfect_enabled 
end

function camera.get_zoom()
    return _zoom
end

function camera.get_target_zoom()
    return _target_zoom
end

function camera.set_movement_smoothing(smoothing)
    check(smoothing, 'smoothing', 'number', 'normalized')
    _movement_smoothing = smoothing
end

function camera.set_zooming_smoothing(smoothing)
    check(smoothing, 'smoothing', 'number', 'normalized')
    _zooming_smoothing = smoothing
end

function camera.set_offset(offset_x, offset_y)
    check(offset_x, 'offset_x', 'number')
    check(offset_y, 'offset_y', 'number')
    _offset_x = offset_x
    _offset_y = offset_y
end

function camera.move_to(x, y)
    check(x, 'x', 'number')
    check(y, 'y', 'number')
    _target_x = x
    _target_y = y
end

function camera.get_position()
    return _x, _y
end

function camera.get_target_position()
    return _target_x, _target_y
end

function camera.zoom_to(zoom_amount)
    check(zoom_amount, 'zoom_amount', 'number', 'strictly_positive')
    _target_zoom = zoom_amount
end

function camera.update()
    if _movement_smoothing > 0 then
        local dx = _target_x - _x
        local dy = _target_y - _y

        if math.abs(dx) > 0.001 then
            _x = _x + dx * _movement_smoothing
        else
            _x = _target_x
        end

        if math.abs(dy) > 0.001 then
            _y = _y + dy * _movement_smoothing
        else
            _y = _target_y
        end
    else
        _x = _target_x
        _y = _target_y
    end

    if _zooming_smoothing > 0 then
        local dzoom = _target_zoom - _zoom

        if math.abs(dzoom) > 0.001 then
            _zoom = _zoom + dzoom * _zooming_smoothing
        else
            _zoom = _target_zoom
        end
    else
        _zoom = _target_zoom
    end

    _transform:reset()
    _transform:translate(_offset_x, _offset_y)
    _transform:scale(_zoom)
    if _pixel_perfect_enabled then
        _transform:translate(-math.floor(_x / _zoom) * _zoom, -math.floor(_y / _zoom) * _zoom)
    else
        _transform:translate(-_x, -_y)
    end
end

function camera.apply_transform()
    love.graphics.applyTransform(_transform)
end

function camera.push_transform()
    love.graphics.push()
    camera.apply_transform()
end

function camera.transform(x, y)
    return _transform:transformPoint(x, y)
end

function camera.inverse_transform(x, y)
    return _transform:inverseTransformPoint(x, y)
end

function camera.is_area_visible(x, y, x2, y2)
    local screen_x, screen_y = camera.transform(x, y)
    local screen_x2, screen_y2 = camera.transform(x2, y2)

    return rectangle.are_intersecting(0,0,viewport.get_width(), viewport.get_height(), screen_x, screen_y, screen_x2, screen_y2)
end

return camera