local pool = require(PATH..'.pool')
local vector = require(PATH..'.vector')
local spatial_hash = require(PATH..'.spatial_hash')
local collision_shapes = require(PATH..'.collision_shapes')

local collider = {}

local _global_hash = spatial_hash.new(100, 100)
local _bodys = {}
local _region_body_map = {}
local _collision_query_pool = {}
local _collision_pool = {}

function collider.get_spatial_hash()
    return _global_hash
end

function collider.new_body(shape, is_dynamic, position_x, position_y)
    local body = {
        shape = shape,
        region_handle = nil,
        is_dynamic = is_dynamic,
        layer = 1,
        mask = {},
        position_x = position_x or 0,
        position_y = position_y or 0,
    }
    local body_handle = tostring(body)

    _bodys[body_handle] = body
    body.region_handle = spatial_hash.insert_region(_global_hash, collision_shapes.get_shape_bounding_box(shape, body.position_x, body.position_y))
    _region_body_map[body.region_handle] = body_handle

    return body_handle
end

function collider.new_dynamic_body(shape, position_x, position_y)
    return collider.new_body(shape, true, position_x, position_y)
end

function collider.new_static_body(shape, position_x, position_y)
    return collider.new_body(shape, false, position_x, position_y)
end

function collider.remove_body(body_handle)
    local body = _bodys[body_handle]
    _bodys[body_handle] = nil
    _region_body_map[body.region_handle] = nil
    spatial_hash.remove_region(_global_hash, body.region_handle)
end

function collider.set_body_collision_layer(body_handle, layer_index)
    local body = _bodys[body_handle]
    body.layer = layer_index
end

function collider.set_body_collision_mask(body_handle, layer_index, state)
    local body = _bodys[body_handle]
    body.mask[layer_index] = state
end

function collider.set_body_position(body_handle, position_x, position_y)
    local body = _bodys[body_handle]
    body.position_x, body.position_y = position_x, position_y
    spatial_hash.set_region(_global_hash, body.region_handle, collision_shapes.get_shape_bounding_box(body.shape, body.position_x, body.position_y))
end

function collider.get_body_position(body_handle)
    local body = _bodys[body_handle]
    return body.position_x, body.position_y
end

function collider.get_body_collisions(body_handle, position_x, position_y)
    local body = _bodys[body_handle]
    local collisions = body.cached_collisions
    local i = 0
    local collisions = pool.fetch(_collision_query_pool)
    local region_x, region_y, region_x2, region_y2 = spatial_hash.get_region(_global_hash, body.region_handle)
    position_x = position_x or body.position_x
    position_y = position_y or body.position_y

    if not (position_x == body.position_x and position_y == body.position_y) then
        local region_displace_x, region_displace_y = vector.add(position_x, position_y, -body.position_x, -body.position_y)
        region_x, region_x2 = vector.translate(region_displace_x, region_x, region_x2)
        region_y, region_y2 = vector.translate(region_displace_y, region_y, region_y2)
    end
    
    for region_handle in spatial_hash.get_overlapping_regions(_global_hash, region_x, region_y, region_x2, region_y2) do
        if region_handle ~= body.region_handle then

            local other_handle = _region_body_map[region_handle]
            local other_body = _bodys[other_handle]

            if not other_body.is_dynamic and body.mask[other_body.layer] ~= false then
                local are_overlapping, separating_vector_x, separating_vector_y = collision_shapes.test_overlap(body.shape, other_body.shape, position_x, position_y, other_body.position_x, other_body.position_y)

                if are_overlapping then
                    local collision = pool.fetch(_collision_pool)
                    collision.body_handle = other_handle
                    collision.overlap_x = separating_vector_x
                    collision.overlap_y = separating_vector_y
                    table.insert(collisions, collision)
                end
            end
        end
    end

    return function()
        i = i + 1
        if i <= #collisions then return collisions[i] end

        for j = #collisions, 1, -1 do
            pool.release(_collision_pool, table.remove(collisions))
        end

        pool.release(_collision_query_pool, collisions)

        return nil
    end
end

function collider.move_and_collide(body_handle, velocity_x, velocity_y)
    for collision in collider.get_body_collisions(body_handle) do

    end
end

return collider