local vector = require(PATH..'.vector')
local collision_shapes = {}

function collision_shapes.point_rectangle(px, py, rx_min, ry_min, rx_max, ry_max)
    local r_width = rx_max - rx_min
    local r_height = ry_max - ry_min
    local x_relative = px - rx_min
    local y_relative = py - ry_min
    local colliding = (x_relative >= 0 and x_relative < r_width) and (y_relative >= 0 and y_relative < r_height)
    local separating_vector_x, separating_vector_y = 0, 0

    if colliding then
        separating_vector_x = -x_relative
        separating_vector_y = -y_relative
    end

    return colliding, separating_vector_x, separating_vector_y
end

function collision_shapes.rectangle_rectangle(ax, ay, ax2, ay2, bx, by, bx2, by2)
    local min_x = math.min(ax, ax2, bx, bx2)
    local max_x = math.max(ax, ax2, bx, bx2)
    local min_y = math.min(ay, ay2, by, by2)
    local max_y = math.max(ay, ay2, by, by2)
    local bounding_box_width = max_x - min_x
    local bounding_box_height = max_y - min_y
    local a_width = math.abs(ax2 - ax)
    local a_height = math.abs(ay2 - ay)
    local b_width = math.abs(bx2 - bx)
    local b_height = math.abs(by2 - by)
    local overlap_width = (a_width + b_width) - bounding_box_width
    local overlap_height = (a_height + b_height) - bounding_box_height
    local colliding = overlap_width > 0 and overlap_height > 0
    local separating_vector_x, separating_vector_y = 0, 0

    if colliding then
        separating_vector_x = overlap_width * (math.min(ax, ax2) < math.min(bx, bx2) and -1 or 1)
        separating_vector_y = overlap_height * (math.min(ay, ay2) < math.min(by, by2) and -1 or 1)
    end

    return colliding, separating_vector_x, separating_vector_y
end

function collision_shapes.circle_rectangle(cx, cy, cr, rx, ry, rx2, ry2)
    local colliding, separating_vector_x, separating_vector_y = collision_shapes.point_rectangle(cx, cy, rx, ry, rx2, ry2)

    if colliding then
        separating_vector_x, separating_vector_y = separating_vector_x-cr, separating_vector_y-cr
    else
        local min_x = math.min(rx, rx2)
        local max_x = math.max(rx, rx2)
        local min_y = math.min(ry, ry2)
        local max_y = math.max(ry, ry2)
        local closest_x = math.min(max_x, math.max(min_x, cx))
        local closest_y = math.min(max_y, math.max(min_y, cy))
        local distance = vector.distance(closest_x, closest_y, cx, cy)
        colliding = distance < cr

        if colliding then
            local overlap_length = cr - distance
            separating_vector_x, separating_vector_y = vector.scale(overlap_length, vector.normalized(cx - closest_x, cy - closest_y))
        end
    end

    return colliding, separating_vector_x, separating_vector_y
end

function collision_shapes.point_circle(px, py, cx, cy, cr)
    return vector.distance(px, py, cx, cy) < cr
end

function collision_shapes.circle_circle(ax, ay, ar, bx, by, br)
    local separating_vector_x, separating_vector_y = 0, 0
    local collision_normal_x, collision_normal_y = 0, 0
    local distance = vector.distance(ax, ay, bx, by)
    local colliding = distance < (ar + br)
    if colliding then
        separating_vector_x, separating_vector_y = vector.scale((ar + br) - distance,vector.normalized(ax - bx, ay - by))
        collision_normal_x, collision_normal_y = vector.normalized(ax - bx, ay - by)
    end
    return colliding, separating_vector_x, separating_vector_y, collision_normal_x, collision_normal_y
end

function collision_shapes.circle_line(cx, cy, cr, lx, ly, lx2, ly2)
    local line_vector_x, line_vector_y = vector.normalized(lx2 - lx, ly2 - ly)
    local circle_center_projection = vector.dot(line_vector_x, line_vector_y, cx-lx, cy-ly)
    local cx_projected, cy_projected = vector.add(lx,ly, vector.scale(circle_center_projection, line_vector_x, line_vector_y))
    cx_projected = math.min(lx2, math.max(lx, cx_projected))
    cy_projected = math.min(ly2, math.max(ly, cy_projected))
    local distance = vector.distance(cx_projected, cy_projected, cx, cy)
    local colliding = distance < cr
    local separating_vector_x, separating_vector_y = 0, 0
    if colliding then
        local overlap_length = cr - distance
        separating_vector_x, separating_vector_y = vector.scale(overlap_length, vector.normalized(cx - cx_projected, cy - cy_projected))
    end
    return colliding, separating_vector_x, separating_vector_y
end

function collision_shapes.circle_capsule(cx, cy, cr, px, py, pw, ph)
    local sub_circle_x, sub_circle_y, sub_circle_radius = px, py, 0.5 * math.min(pw, ph)

    if pw > ph then
        local half_width = 0.5 * pw
        sub_circle_x = math.min(px + half_width - sub_circle_radius, math.max(px - half_width + sub_circle_radius, cx))
    elseif pw < ph then
        local half_height = 0.5 * ph
        sub_circle_y = math.min(py + half_height - sub_circle_radius, math.max(py - half_height + sub_circle_radius, cy))
    end

    return collision_shapes.circle_circle(cx, cy, cr, sub_circle_x, sub_circle_y, sub_circle_radius)
end

function collision_shapes.capsule_capsule(ax, ay, aw, ah, bx, by, bw, bh)
    local a_sub_circle_x, a_sub_circle_y = ax, ay
    local a_radius = 0.5 * math.min(aw, ah)
    local b_radius = 0.5 * math.min(bw, bh)
    --local b_sub_circle_x, b_sub_circle_y, b_sub_circle_radius = bx, by, 0.5 * math.min(bw, bh)

    if aw > ah and bw > bh then -- a horizontal and a // b
        local a_half_width = 0.5* aw - a_radius
        local a_left = ax - a_half_width
        local a_right = ax + a_half_width
        local b_half_width = 0.5 * bw - b_radius
        local left = math.min(bx + b_half_width, math.max(bx - b_half_width, a_left))
        local right = math.min(bx + b_half_width, math.max(bx - b_half_width, a_right))
        if math.abs(a_left - left) < math.abs(a_right - right) then
            a_sub_circle_x = a_left
        else
            a_sub_circle_x = a_right
        end
    elseif aw < ah and bw < bh then  -- a vertical and a // b
        local a_half_height = 0.5*(aw-2*a_radius)
        local a_top = ay - a_half_height
        local a_bottom = ay + a_half_height

        local b_half_height = 0.5 * bw - b_radius
        local top = math.min(by + b_half_height, math.max(by - b_half_height, a_top))
        local bottom = math.min(by + b_half_height, math.max(by - b_half_height, a_bottom))
        if math.abs(a_top - top) < math.abs(a_bottom - bottom) then
            a_sub_circle_y = top
        else
            a_sub_circle_y = bottom
        end
    elseif aw > ah then -- a horizontal and a -|- b
        local a_half_width = 0.5*(aw-2*a_radius)
        a_sub_circle_x = math.min(ax + a_half_width, math.max(ax - a_half_width, bx))
    else -- a vertical and a -|- b
        local a_half_height = 0.5*(ah-2*a_radius)
        a_sub_circle_y = math.min(ay + a_half_height, math.max(ay - a_half_height, by))
    end

    return collision_shapes.circle_capsule(a_sub_circle_x, a_sub_circle_y, a_radius, bx, by, bw, bh)
end

local function get_good_axes(vertexes, axes)
    local vertex_count = 0.5 * #vertexes

    for i = 1, vertex_count-1 do
        local vertex_x, vertex_y = vertexes[i*2-1], vertexes[i*2]
        local next_vertex_index = i % vertex_count + 1
        local next_vertex_x, next_vertex_y = vertexes[next_vertex_index*2-1], vertexes[next_vertex_index*2]
        local edge_normal_x, edge_normal_y = vector.normal(next_vertex_x - vertex_x, next_vertex_y - vertex_y)
        if axes[tostring(edge_normal_x)..','..tostring(edge_normal_y)] == nil and axes[tostring(-edge_normal_y)..','..tostring(-edge_normal_x)] == nil then
            axes[tostring(edge_normal_x)..','..tostring(edge_normal_y)] = true
        end
    end
end

local function get_projection(vertexes, axis_x, axis_y)
    local min_projection, max_projection
    local vertex_count = 0.5 * #vertexes

    for i = 1, vertex_count do
        local vertex_x, vertex_y = vertexes[i*2-1], vertexes[i*2]
        local vertex_projection = vector.dot(vertex_x, vertex_y, axis_x, axis_y)

        if i == 1 then
            min_projection, max_projection = vertex_projection, vertex_projection
        else
            min_projection = math.min(min_projection, vertex_projection)
            max_projection = math.max(max_projection, vertex_projection)
        end
    end

    return min_projection, max_projection
end

local function sat_test(a_vertexes, b_vertexes)
    local a_vertex_count, b_vertex_count = 0.5 * #a_vertexes, 0.5 * #b_vertexes
    local axes = {}
    local separated = false
    local separating_vector_x, separating_vector_y = 0, 0
    local min_overlap_length

    get_good_axes(a_vertexes, axes)
    get_good_axes(b_vertexes, axes)

    for axis in pairs(axes) do
        local axis_x, axis_y = axis:match('(%.),(%.)')
        axis_x = tonumber(axis_x)
        axis_y = tonumber(axis_y)
        local a_min_projection, a_max_projection = get_projection(a_vertexes, axis_x, axis_y)
        local b_min_projection, b_max_projection = get_projection(b_vertexes, axis_x, axis_y)
        local a_projection_width, b_projection_width = a_max_projection - a_min_projection, b_max_projection - b_min_projection
        local projection_width = math.max(a_max_projection, b_max_projection) - math.min(a_min_projection, b_min_projection)
        local projection_overlap = a_projection_width + b_projection_width - projection_width

        if projection_overlap <= 0 then 
            separated = true
            separating_vector_x, separating_vector_y = 0, 0
            break
        else
            if min_overlap_length == nil or projection_overlap < min_overlap_length then
                min_overlap_length = projection_overlap
                local axis_angle = vector.angle_between(axis_x, axis_y, 1, 0)
                if a_min_projection < b_min_projection then
                    separating_vector_x, separating_vector_y = vector.rotate(axis_angle, -projection_overlap, 0)
                else
                    separating_vector_x, separating_vector_y = vector.rotate(axis_angle, projection_overlap, 0)
                end
            end
        end
    end

    return separated, separating_vector_x, separating_vector_y
end

function collision_shapes.convex_convex(polygon_a_vertexes, polygon_b_vertexes)
    return sat_test(polygon_a_vertexes, polygon_b_vertexes)
end

function collision_shapes.new_shape(shape_name)
    return { _shape_name = shape_name }
end

function collision_shapes.new_circle(radius, center_x, center_y)
    local circle = collision_shapes.new_shape('circle')
    circle.center_x = center_x or 0
    circle.center_y = center_y or 0
    circle.radius = radius
    return circle
end

function collision_shapes.new_line(x, y, x2, y2)
    local line = collision_shapes.new_shape('line')
    line.x = x
    line.y = y
    line.x2 = x2
    line.y2 = y2
    return line
end

function collision_shapes.new_rectangle(x, y, x2, y2)
    local rectangle = collision_shapes.new_line(math.min(x , x2), math.min(y, y2), math.max(x, x2), math.max(y, y2))
    rectangle._shape_name = 'rectangle'
    return rectangle
end

function collision_shapes.new_capsule(width, height, x, y)
    local capsule = collision_shapes.new_shape('capsule')
    capsule.x = x or 0
    capsule.y = y or 0
    capsule.width = width
    capsule.height = height
    return capsule
end

function collision_shapes.get_shape_bounding_box(shape, offset_x, offset_y)
    local x, y, x2, y2 = 0, 0, 0, 0
    offset_x = offset_x or 0
    offset_y = offset_y or 0

    if shape._shape_name == 'circle' then
        x, y, x2, y2 = shape.center_x - shape.radius, shape.center_y - shape.radius, shape.center_x + shape.radius, shape.center_y + shape.radius
    elseif shape._shape_name == 'convex' then
        x, y, x2, y2 = math.min(unpack(shape.vertexes_x)), math.min(unpack(shape.vertexes_y)), math.max(unpack(shape.vertexes_x)), math.max(unpack(shape.vertexes_y))
    elseif shape._shape_name == 'capsule' then
        x, y, x2, y2 = shape.x - 0.5 * shape.width, shape.y - 0.5 * shape.height, shape.x + 0.5 * shape.width, shape.y + 0.5 * shape.height
    else
        x, y, x2, y2 = shape.x, shape.y, shape.x2, shape.y2
    end

    return x + offset_x, y + offset_y, x2 + offset_x, y2 + offset_y
end

function collision_shapes.test_overlap(shape_a, shape_b, a_translation_x, a_translation_y, b_translation_x, b_translation_y)
    local are_overlapping, separating_vector_x, separating_vector_y = false, 0, 0
    local shape_name_a, shape_name_b = shape_a._shape_name, shape_b._shape_name

    if shape_name_a == 'rectangle' and shape_name_b == 'circle' then
        shape_name_a, shape_name_b = shape_name_b, shape_name_a
    end

    if shape_name_a == 'circle' and shape_name_b == 'circle' then
        are_overlapping, separating_vector_x, separating_vector_y = collision_shapes.circle_circle(
            shape_a.center_x + a_translation_x, 
            shape_a.center_y + a_translation_y, 
            shape_a.radius, 
            shape_b.center_x + b_translation_x, 
            shape_b.center_y + b_translation_y, 
            shape_b.radius)
    elseif shape_name_a == 'circle' and shape_name_b == 'rectangle' then
        are_overlapping, separating_vector_x, separating_vector_y = collision_shapes.circle_rectangle(
            shape_a.center_x + a_translation_x, 
            shape_a.center_y + a_translation_y, 
            shape_a.radius, 
            shape_b.x + b_translation_x, 
            shape_b.y + b_translation_y, 
            shape_b.x2 + b_translation_x, 
            shape_b.y2 + b_translation_y)
    end

    return are_overlapping, separating_vector_x, separating_vector_y
end

function collision_shapes.draw(shape, drawmode)
    drawmode = drawmode or "line"
    love.graphics.setLineWidth(2)
    if shape._shape_name == 'circle' then
        love.graphics.circle(drawmode,shape.center_x,shape.center_y,shape.radius)
    elseif shape._shape_name == 'rectangle' then
        love.graphics.rectangle(drawmode,shape.x,shape.y,shape.x2-shape.x,shape.y2-shape.y)
    elseif shape._shape_name == 'capsule' then
        local radius = 0.5*math.min(shape.width,shape.height)
        love.graphics.rectangle(drawmode,
            shape.x-0.5*shape.width,
            shape.y-0.5*shape.height,
            shape.width,shape.height,radius)
    end
end

return collision_shapes