local color = {}

local max, min, abs = math.max, math.min, math.abs
local floor = math.floor

local HEX_DIGITS = {0,1,2,3,4,5,6,7,8,9,'a','b','c','d','e','f'}

local function round(x) return math.floor(x + 0.5) end

function color.rgb_to_hex(r, g, b, a)
    local s = 1/16
    local d = HEX_DIGITS
    r, g, b, a = floor(r*255), floor(g*255), floor(b*255), floor(a*255)
    return d[floor(r * s)+1]..d[r % 16+1]..d[floor(g * s)+1]..d[g % 16+1]..d[floor(b * s)+1]..d[b % 16+1]..d[floor(a * s)+1]..d[a % 16+1]
end

function color.rgb_to_hsv(r, g, b, a)
    local max, min = max(r,g,b), min(r,g,b)
    local c = max - min
    local t, u = 0.5 * (2*r - g - b), 0.866 * (g - b)

    return round(math.deg(math.atan2(u, t))), 100 * (max > 0 and c / max or 0), 100 * max, a
end

function color.rgb_to_hsl(r, g, b, a)
    local max, min = max(r,g,b), min(r,g,b)
    local c = max - min
    local t, u = 0.5 * (2*r - g - b), 0.866 * (g - b)
    local l = 0.5 * (max + min)

    return round(math.deg(math.atan2(u, t))), 100 * ((l > 0 or l < 1) and c / (1 - math.abs(2*l - 1)) or 0), 100 * l, a
end

function color.hsv_to_hsl(h, s, v, a)
    s, v = 0.01 * s, 0.01 * v
    local l = v * (1 - 0.5*s)
    return h, 100 * ((l == 0 or l == 1) and 0 or (v-l)/(min(l,1-l))), 100 * l, a
end

function color.hsl_to_hsv(h, s, l, a)
    s, l = 0.01 * s, 0.01 * l
    local v = l + s * min(l,1-l)
    return h, 100 * ((v == 0) and 0 or 2 * (1 - l/v)), 100 * v, a
end

function color.hex_to_rgb(hex_string)
    local code = string.match(string.lower(hex_string), '#*([a-f0-9]+)')
    local c
    local r_string, g_string, b_string, a_string = '0','0','0','f'
    local code_length = string.len(code)
    local sub = string.sub

    if not(code_length == 6 or code_length == 8 or code_length == 4 or code_length == 3) then error('hex_to_rgb: invalid input string') end

    if code_length == 6 or code_length == 8 then
        c = 1/255
        r_string, g_string, b_string, a_string  = sub(code, 1,2), sub(code, 3,4), sub(code, 5,6), code_length == 8 and sub(code, 7,8) or 'ff'
    else
        c = 1/15
        r_string, g_string, b_string, a_string  = sub(code, 1,1), sub(code, 2,2), sub(code, 3,3), code_length == 4 and sub(code, 4,4) or 'f'
    end

    return c * tonumber('0x'..r_string), c * tonumber('0x'..g_string), c * tonumber('0x'..b_string), c * tonumber('0x'..a_string)
end

function color.hsl_to_rgb(h, s, l, a)
    l = l * 0.01
    if s<=0 then return l,l,l,a end
    h, s = h/360*6, s * 0.01
    local c = (1-abs(2*l-1))*s
    local x = (1-abs(h%2-1))*c
    local m,r,g,b = (l-0.5*c), 0,0,0
    if h < 1     then r,g,b = c,x,0
    elseif h < 2 then r,g,b = x,c,0
    elseif h < 3 then r,g,b = 0,c,x
    elseif h < 4 then r,g,b = 0,x,c
    elseif h < 5 then r,g,b = x,0,c
    else              r,g,b = c,0,x
    end return (r+m),(g+m),(b+m),a or 1
end

function color.hsv_to_rgb(h, s, v, a)
    return color.hsl_to_rgb(color.hsv_hsl(h, s, v, a))
end

function color.grayscale(v, a) return v, v, v, a or 1 end

function color.new_hex_color(hex_string) return {color.hex_to_rgb(hex_string)} end
function color.new_hsl_color(h,s,l,a) return {color.hsl_to_rgb(h,s,l,a)} end
function color.new_hsv_color(h,s,v,a) return {color.hex_to_rgb(h,s,v,a)} end
function color.new_grayscale_color(v,a) return {color.grayscale(v,a)} end

color.transparent = {1,1,1,0}
color.red = {1,0,0,1}
color.orange = {1,0.5,0,1}
color.yellow = {1,1,0,1}
color.lime = {0.5,1,0,1}
color.green = {0,1,0,1}
color.aqua = {0,1,1,1}
color.teal = {0,0.5,1,1}
color.blue = {0,0,1,1}
color.magenta = {1,0,1,1}
color.white = {1,1,1,1}
color.black = {0,0,0,1}
color.gray = {0.5,0.5,0.5,1}
color.lightgray = {0.75,0.75,0.75,1}

return color