local rectangle = require(PATH..'.rectangle')
local pool = require(PATH..'.pool')

local box = {
    is_debug = false,
}
local box_metatable = {}
box_metatable.__index = box_metatable

local traversal_pool = {}

function box.new_box()
    local b = {
        width=1,height=1,
        visible_region={0,0,1,1},
        children={},
        local_transform=love.math.newTransform(),
        layout_transform=love.math.newTransform(),
        global_transform=love.math.newTransform(),
        z = 0,
        scroll_x=0,
        scroll_y=0,
        scroll_max_x=0,
        scroll_max_y=0,
        stretch_basis_x=1,
        stretch_max_width=nil,
        stretch_min_width=nil,
        stretch_x=nil,
        stretch_basis_y=1,
        stretch_max_height=nil,
        stretch_min_height=nil,
        stretch_y=nil,
        parent=nil,
        draw_method=nil,
        draw_options = {},
        layout_method = nil,
        layout_options = {},
        is_debug = true,
    }
    setmetatable(b, box_metatable)
    return b
end

function box.sort_by_z(box_a, box_b)
    if box_a.z == box_b.z then
        return tostring(box_a) < tostring(box_b)
    else
        return box_a.z < box_b.z
    end
end

function box.insert_box(child, parent, index)
    child.parent = parent
    table.insert(parent.children, index, child)
end

function box.append_box(child, parent)
    box.insert_box(child,parent,#parent.children+1)
end

function box.remove_box(child)
    if child.parent then
        local siblings=child.parent.children
            for i=1, #siblings do
                if siblings[i] == child then
                table.remove(siblings, i)
                break
            end
        end
        child.parent = nil
    end
end

function box.add_scroll(box, dx, dy)
    box.scroll_x=math.min(box.scroll_max_x, math.max(0, box.scroll_x-dx))
    box.scroll_y=math.min(box.scroll_max_y, math.max(0, box.scroll_y-dy))
end

function box.update_visible_region(box_tree)
    if box_tree.parent then
        local parent = box_tree.parent
        local box_left, box_top = box_tree.global_transform:transformPoint(0,0)
        local box_right, box_bottom = box_tree.global_transform:transformPoint(box_tree.width, box_tree.height)
        rectangle.copy(box_tree.visible_region, rectangle.intersection(box_left,box_top,box_right,box_bottom,unpack(parent.visible_region)))
    else
        box_tree.visible_region[1], box_tree.visible_region[2] = box_tree.global_transform:transformPoint(0,0)
        box_tree.visible_region[3], box_tree.visible_region[4] = box_tree.global_transform:transformPoint(box_tree.width, box_tree.height)
    end
end

function box.is_box_visible(box)
    return not rectangle.is_null(unpack(box.visible_region))
end

function box.is_point_inside_box(box, x, y)
    local local_x, local_y = box.global_transform:inverseTransformPoint(x,y)
    return rectangle.is_point_inside(local_x, local_y, 0, 0, box.width, box.height)
end

function box.update_global_transform(b)
    b.global_transform:reset()
    b.global_transform:apply(b.layout_transform)
    b.global_transform:apply(b.local_transform)
end

function box.layout_box(box)
    if box.layout_method then
        box:layout_method()
    end
end

function box.layout_box_tree(box_tree)
    local traversal_table = pool.fetch(traversal_pool)
    local traversal_index=1

    traversal_table[1] = box_tree

    while traversal_index<=#traversal_table do
        local current_box = traversal_table[traversal_index]
        for i=1, #current_box.children do
            table.insert(traversal_table, current_box.children[i])
        end
        traversal_index=traversal_index+1
    end

    for i, current_box in ipairs(traversal_table) do
        box.layout_box(current_box)
    end

    pool.release(traversal_pool, traversal_table)
end

function box.get_box_at(box_tree, global_x, global_y)
    local tested_set = pool.fetch(traversal_pool)
    local hit_box = nil

    tested_set[1] = box_tree

    while #tested_set > 0 do
        local hit_box_in_set

        for i = #tested_set, 1, -1 do
            local tested_box = tested_set[i]

            if box.is_point_inside_box(tested_box, global_x, global_y) then
                hit_box_in_set = tested_box

                break
            end
        end

        for j = #tested_set, 1, -1 do
            table.remove(tested_set, j)
        end

        if hit_box_in_set then
            hit_box = hit_box_in_set

            for j = 1, #hit_box_in_set.children do
                table.insert(tested_set, hit_box_in_set.children[j])
            end

            if hit_box_in_set.ordering_method then
                table.sort(tested_set, hit_box_in_set.ordering_method)
            else
                table.sort(tested_set, box.sort_by_z)
            end
        end
    end

    return hit_box
end

function box.draw_box(box_tree)
    --if is_box_visible(box) then
    love.graphics.push()
    --love.graphics.setScissor(box.visible_region.x,box.visible_region.y,box.visible_region.width,box.visible_region.height)
    love.graphics.applyTransform(box_tree.global_transform)
    if box_tree.draw_method then
        box_tree:draw_method()
    elseif box.is_debug then
        love.graphics.setColor(1,0.3,0.7)
        love.graphics.setLineWidth(4)
        love.graphics.rectangle("line",0,0,box_tree.width,box_tree.height,10)
    end
    love.graphics.pop()
    --end
end

function box.draw_box_tree(box_tree)
    if not box_tree.is_scissor_enabled or box.is_box_visible(box_tree) then
        love.graphics.push()
        love.graphics.applyTransform(box_tree.global_transform)

        if box_tree.is_scissor_enabled and box_tree.parent then
            local sx, sy, sw, sh = rectangle.to_box(unpack(box_tree.parent.visible_region))
            love.graphics.setScissor(sx, sy, love.graphics.transformPoint(sw, sh))
        end

        if box_tree.draw_method then
            box_tree:draw_method()
        elseif box.is_debug and box_tree.is_debug then
            love.graphics.setColor(1,0.3,0.7)
            love.graphics.setLineWidth(4)
            love.graphics.rectangle("line",0,0,box_tree.width,box_tree.height,10)
        end
        love.graphics.setScissor()
        love.graphics.pop()

        local children_sequence = pool.fetch(traversal_pool)
        
        for i, child in ipairs(box_tree.children) do
            children_sequence[i] = child
        end

        if box_tree.ordering_method then
            table.sort(children_sequence, box_tree.ordering_method)
        else
            table.sort(children_sequence, box.sort_by_z)
        end

        for i, child in ipairs(children_sequence) do
            child:draw_tree()
        end

        pool.release(traversal_pool, children_sequence)
    end
end

box_metatable.insert = box.insert_box
box_metatable.append = box.append_box
box_metatable.remove = box.remove_box
box_metatable.draw = box.draw_box
box_metatable.draw_tree = box.draw_box_tree
box_metatable.layout = box.layout_box
box_metatable.layout_tree = box.layout_box_tree
box_metatable.is_point_inside = box.is_point_inside_box
box_metatable.is_visible = box.is_box_visible
box_metatable.add_scroll = box.add_scroll
box_metatable.update_visible_region = box.update_visible_region
box_metatable.get_box_at = box.get_box_at
box_metatable.update_global_transform = box.update_global_transform

return box