local rectangle = require(GK_PATH..'.rectangle')
local pool = require(GK_PATH..'.pool')

local box = {
    draw_wireframe_enabled = false,
}

local Box = {}
Box.__index = Box

local traversal_pool = {}

function box.new_box()
    local b = {
        width=1,height=1,
        visible_region={0,0,1,1},
        children={},
        local_transform=love.math.newTransform(),
        layout_transform=love.math.newTransform(),
        global_transform=love.math.newTransform(),
        z = 0,
        scroll_x=0,
        scroll_y=0,
        scroll_max_x=0,
        scroll_max_y=0,
        stretch_basis_x=1,
        stretch_max_width=nil,
        stretch_min_width=nil,
        stretch_x=nil,
        stretch_basis_y=1,
        stretch_max_height=nil,
        stretch_min_height=nil,
        stretch_y=nil,
        parent=nil,
        draw_method=nil,
        draw_options = {},
        layout_method = nil,
        layout_options = {},
        is_debug = true,
    }
    setmetatable(b, Box)
    return b
end

function box.sort_by_z(box_a, box_b)
    if box_a.z == box_b.z then
        return tostring(box_a) < tostring(box_b)
    else
        return box_a.z < box_b.z
    end
end

function Box:insert(parent, index)
    self:remove()

    self.parent = parent
    table.insert(parent.children, index, self)
end

function Box:append(parent)
    self:insert(parent, #parent.children+1)
end

function Box:remove()
    if self.parent then
        local siblings = self.parent.children
            for i=1, #siblings do
                if siblings[i] == self then
                table.remove(siblings, i)
                break
            end
        end
        self.parent = nil
    end
end

function Box:add_scroll(dx, dy)
    self.scroll_x=math.min(self.scroll_max_x, math.max(0, self.scroll_x-dx))
    self.scroll_y=math.min(self.scroll_max_y, math.max(0, self.scroll_y-dy))
end

function Box:update_visible_region()
    if self.parent then
        local parent = self.parent
        local box_left, box_top = self.global_transform:transformPoint(0,0)
        local box_right, box_bottom = self.global_transform:transformPoint(self.width, self.height)
        rectangle.copy(self.visible_region, rectangle.intersection(box_left,box_top,box_right,box_bottom,unpack(parent.visible_region)))
    else
        self.visible_region[1], self.visible_region[2] = self.global_transform:transformPoint(0,0)
        self.visible_region[3], self.visible_region[4] = self.global_transform:transformPoint(self.width, self.height)
    end
end

function Box:is_visible()
    return not rectangle.is_null(unpack(self.visible_region))
end

function Box:is_point_inside(global_x, global_y)
    local local_x, local_y = self.global_transform:inverseTransformPoint(global_x, global_y)
    return rectangle.is_point_inside(local_x, local_y, 0, 0, self.width, self.height)
end

function Box:update_global_transform()
    self.global_transform:reset()
    self.global_transform:apply(self.layout_transform)
    self.global_transform:apply(self.local_transform)
end

function Box:layout()
    local traversal_table = pool.fetch(traversal_pool)
    local traversal_index=1

    traversal_table[1] = self

    while traversal_index<=#traversal_table do
        local current_box = traversal_table[traversal_index]
        for i=1, #current_box.children do
            table.insert(traversal_table, current_box.children[i])
        end
        traversal_index=traversal_index+1
    end

    for i, current_box in ipairs(traversal_table) do
        if current_box.layout_method then
            current_box:layout_method()
        else
            current_box:update_visible_region()
        end
    end

    pool.release(traversal_pool, traversal_table)
end

function Box:get_box_at(global_x, global_y, check_only_subtree)
    local tested_set = pool.fetch(traversal_pool)
    local hit_box = nil

    if check_only_subtree == true then
        tested_set[1] = self
    else
        local root = self
        while root.parent do
            root = root.parent
        end
        tested_set[1] = root
    end

    while #tested_set > 0 do
        local hit_box_in_set

        for i = #tested_set, 1, -1 do
            local tested_box = tested_set[i]

            if tested_box:is_visible() and tested_box:is_point_inside(global_x, global_y) then
                hit_box_in_set = tested_box

                break
            end
        end

        for j = #tested_set, 1, -1 do
            table.remove(tested_set, j)
        end

        if hit_box_in_set then
            hit_box = hit_box_in_set

            for j = 1, #hit_box_in_set.children do
                table.insert(tested_set, hit_box_in_set.children[j])
            end

            if hit_box_in_set.ordering_method then
                table.sort(tested_set, hit_box_in_set.ordering_method)
            else
                table.sort(tested_set, box.sort_by_z)
            end
        end
    end

    return hit_box
end

function Box:is_box_at(global_x, global_y, check_only_subtree)
    return self == self:get_box_at(global_x, global_y, check_only_subtree)
end

function Box:draw()
    if self:is_visible() then
        love.graphics.push()
        love.graphics.applyTransform(self.global_transform)

        if self.is_scissor_enabled then
            love.graphics.setScissor(rectangle.to_box(unpack(self.visible_region)))
        end

        if self.draw_method then
            self:draw_method()

        elseif box.draw_wireframe_enabled then
            love.graphics.setColor(1,0.3,0.7)
            love.graphics.setLineWidth(2)
            love.graphics.rectangle("line",0,0,self.width,self.height,10)
        end

        love.graphics.setScissor()
        love.graphics.pop()

        local children_sequence = pool.fetch(traversal_pool)
        
        for i, child in ipairs(self.children) do
            children_sequence[i] = child
        end

        if self.ordering_method then
            table.sort(children_sequence, self.ordering_method)
        else
            table.sort(children_sequence, box.sort_by_z)
        end

        for i, child in ipairs(children_sequence) do
            child:draw()
        end

        pool.release(traversal_pool, children_sequence)
    end
end

return box