local utf8 = require("utf8")
local font = require(PATH..'.font')

local string_len = string.len
local string_sub = string.sub
local utf_len = utf8.len

local text_input = {}
local SingleLineInput = {}
local MultiLineInput = {}
SingleLineInput.__index = SingleLineInput
MultiLineInput.__index = MultiLineInput

function text_input.new_single_line_input()
    local input = {
        text = "",
        carriage_index = 1,
        font = font.love(20),
        selection_left_character = nil,
        selection_right_character = nil,
    }
    setmetatable(input, SingleLineInput)
    return input
end

function text_input.new_multi_line_input()
    local input = text_input.new_single_line()
    input.lines_table = {{start_index = 1, end_index = 1}}
    input.selection_left_line = nil
    input.selection_right_line = nil
    setmetatable(input, MultiLineInput)
    return input
end

function SingleLineInput:get_character_at_position(local_x, local_y)
    local text_width = self.font:getWidth(self.text)
    local character_index
    
    if local_x <= 0 then
        character_index = 0
    elseif local_x >= text_width then
        character_index = string_len(self.text)
    else
        for i = 1, string_len(self.text) do
            local text_so_far = string_sub(self.text, 1,i-1)
            local character = string_sub(self.text, i,i)

            if self.font:getWidth(text_so_far) + 0.5 * self.font:getWidth(character) > local_x then
                character_index = i-1
                break
            end
            if character_index == nil and i == string_len(self.text) then
                character_index = string_len(self.text)
            end
        end
    end

    return character_index
end

function SingleLineInput:insert_text(new_substring)
    if utf_len(new_substring) ~= string_len(new_substring) then return end

    local old_left, old_right
    local new_text
    local new_carriage_index

    if self.selection_left_character and self.selection_right_character then
        old_left = string_sub(self.text, 1, self.selection_left_character-1)
        old_right = string_sub(self.text, self.selection_right_character + 1)
    else
        old_left = string_sub(self.text, 1, self.carriage_index)
        old_right = string_sub(self.text, self.carriage_index + 1)
    end

    new_text = old_left .. new_substring .. old_right
    new_carriage_index = string.len(old_left .. new_substring)

    self.text = new_text
    self.carriage_index = new_carriage_index
    text_input.deselect(self)
end

function SingleLineInput:select_range(left_index, right_index)
    self.selection_left_character = math.max(1, left_index)
    self.selection_right_character = math.min(string_len(self.text), right_index)
end

function SingleLineInput:has_selection()
    return self.selection_left_character and self.selection_right_character
end

function SingleLineInput:get_selected_text()
    if text_input.has_selection(self) then
        return string_sub(self.text, self.selection_left_character, self.selection_right_character)
    else
        return ""
    end
end

function SingleLineInput:deselect()
    self.selection_left_character = nil
    self.selection_right_character = nil
end

function SingleLineInput:set_carriage_at(character_index)
    self.carriage_index = math.min(string_len(self.text), math.max(0, character_index))
end

function SingleLineInput:set_text(text)
    self.text = text
    self:deselect()
end

function SingleLineInput:get_text()
    return self.text
end

return text_input