local pool = require(PATH..'.pool')
local input = {}

local frame_input = {}
local user_controls = {}

local event_entry_pool = {}

function input.map_event_to_control(control_name, device_name, ...)
    if not user_controls[control_name] then
        user_controls[control_name] = {}
    end
    for i = 1, 6 do
        table.insert(user_controls[control_name], { device_name = device_name, ... })
    end
end

function input.map_controls(map_table)
    for control_name, map in pairs(map_table) do
        for i, map_entry in ipairs(map) do
            input.map_event_to_control(control_name, unpack(map_entry))
        end
    end
end

input.map_event_to_control("ctrl", "key", "lctrl")
input.map_event_to_control("ctrl", "key", "rctrl")

function input.match_event(input_name, ...)
    local result = false
    local tested_arguments_count = select('#', ...)
    local frame_input_events = frame_input[input_name]

    if frame_input_events then
        for i, event in ipairs(frame_input_events) do
            if not result then
                local is_event_matching
                for j = 1, tested_arguments_count do
                    local queried_value = select(j, ...)
                    
                    if is_event_matching == nil then
                        is_event_matching = (queried_value == nil or event[j] == queried_value)
                    else
                        is_event_matching = is_event_matching and (queried_value == nil or event[j] == queried_value)
                    end
                end
                result = is_event_matching
            end
        end
    end
    return result
end

function input.is_key_pressed(key, scancode)
    return input.match_event('keypressed', key, scancode)
end

function input.is_key_released(key, scancode)
    return input.match_event('keyreleased', key, scancode)
end

function input.is_key_down(key, scancode)
    return key ~= nil and love.keyboard.isDown(key) or scancode ~= nil and love.keyboard.isScancodeDown(scancode)
end

function input.is_mouse_pressed(button)
    return input.match_event('mousepressed', nil,nil,button)
end

function input.is_mouse_released(button)
    return input.match_event('mousereleased', nil,nil,button)
end

function input.is_mouse_down(button, scancode)
    return button ~= nil and love.mouse.isDown(button)
end

function input.is_control_pressed(control_name)
    local control = user_controls[control_name]

    if control then
        for i, entry in ipairs(control) do
            local handler_name = 'is_'..entry.device_name..'_pressed'
            if input[handler_name](unpack(entry)) then
                return true
            end
        end
    end

    return false
end

function input.is_control_released(control_name)
    local control = user_controls[control_name]

    if control then
        for i, entry in ipairs(control) do
            local handler_name = 'is_'..entry.device_name..'_released'
            if input[handler_name](unpack(entry)) then
                return true
            end
        end
    end

    return false
end

function input.is_control_down(control_name)
    local control = user_controls[control_name]

    if control then
        for i, entry in ipairs(control) do
            local handler_name = 'is_'..entry.device_name..'_down'
            if input[handler_name](unpack(entry)) then
                return true
            end
        end
    end

    return false
end

function input.poll()
    local i = 1
    local event_count = #frame_input

    return function()
        if i > event_count then return nil end

        local event = frame_input[i]

        i = i + 1

        return event.input_name, unpack(event)
    end
end

function input.handle(handler_function)
    local catched_events = pool.fetch(event_entry_pool)

    for i = 1, #frame_input do
        local event = frame_input[i]
        
        if handler_function(event.input_name, unpack(event)) then
            table.insert(catched_events, i)
        end
    end

    for i = 1, #catched_events do
        local event = table.remove(frame_input, catched_events[i])
        local event_siblings = frame_input[event.input_name]

        for i = 1, #event_siblings do
            if event_siblings[i] == event then
                table.remove(event_siblings, i)
                break
            end
        end
    end

    pool.release(event_entry_pool, catched_events)
end

local function clear_previous_input()
    for i = #frame_input, 1, -1 do
        table.remove(frame_input)
    end
    for input_name, events in pairs(frame_input) do
        for i = #events, 1, -1 do
            pool.release(event_entry_pool, table.remove(events))
        end
    end
end

local function input_handle(input_name, ...)
    if not frame_input[input_name] then
        frame_input[input_name] = {}
    end
    local event_attributes = pool.fetch(event_entry_pool)
    event_attributes.input_name = input_name
    for i = 1, 6 do
        event_attributes[i] = select(i, ...)
    end
    table.insert(frame_input, event_attributes)
    table.insert(frame_input[input_name], event_attributes)
end

function love.run()
	if love.load then love.load(love.arg.parseGameArguments(arg), arg) end

	-- We don't want the first frame's dt to include time taken by love.load.
	if love.timer then love.timer.step() end

	local dt = 0

	-- Main loop time.
	return function()
		-- Process events.
		if love.event then
            clear_previous_input()
			love.event.pump()
			for name, a,b,c,d,e,f in love.event.poll() do
				if name == "quit" then
					if not love.quit or not love.quit() then
						return a or 0
					end
				end
                input_handle(name, a, b, c, d, e, f)
				love.handlers[name](a,b,c,d,e,f)
			end
		end

		-- Update dt, as we'll be passing it to update
		if love.timer then dt = love.timer.step() end

		-- Call update and draw
		if love.update then love.update(dt) end -- will pass 0 if love.timer is disabled

		if love.graphics and love.graphics.isActive() then
			love.graphics.origin()
			love.graphics.clear(love.graphics.getBackgroundColor())

			if love.draw then love.draw() end

			love.graphics.present()
		end

		if love.timer then love.timer.sleep(0.001) end
	end
end

return input