local vector = require(PATH..'.vector')

local ruler = {}

function ruler.get_grid_cell_at(x, y, cell_width, cell_height, grid_offset_x, grid_offset_y)
    grid_offset_x = grid_offset_x or 0
    grid_offset_y = grid_offset_y or 0

    return math.floor((x - grid_offset_x) / cell_width), math.floor((y - grid_offset_y) / cell_height)
end

function ruler.grid_snap_position(x, y, cell_width, cell_height, grid_offset_x, grid_offset_y)
    return vector.multiply(cell_width, cell_height, ruler.get_grid_cell_at(x, y, cell_width, cell_height, grid_offset_x, grid_offset_y))
end

function ruler.get_relative_to_range(ratio, range_min, range_max)
    return range_min + math.floor((range_max - range_min) * ratio + 0.5)
end

function ruler.align_box_around_position(box_width, box_height, position_x, position_y, horizontal_alignment, vertical_alignment, offset_x, offset_y)
    offset_x = offset_x or 0
    offset_y = offset_y or 0
    local aligned_x, aligned_y = position_x + offset_x, position_y + offset_y

    if horizontal_alignment == 'center' then
        aligned_x = position_x - ruler.get_relative_to_range(0.5, 0, box_width)
    elseif horizontal_alignment == 'right' then
        aligned_x = position_x - (box_width + offset_x)
    end

    if vertical_alignment == 'center' then
        aligned_y = position_y - ruler.get_relative_to_range(0.5, 0, box_height)
    elseif vertical_alignment == 'bottom' then
        aligned_y = position_y - (box_height + offset_y)
    end

    return aligned_x, aligned_y
end

function ruler.align_box_around_box(box_width, box_height, other_x, other_y, other_width, other_height, side, margin)
    margin = margin or 0
    local box_x, box_y = other_x, other_y + other_width + margin

    if side == 'top' then
        box_y = other_y - box_height - margin
    elseif side == 'left' then
        box_x = other_x - box_width - margin
        box_y = other_y
    elseif side == 'right' then
        box_x = other_x + box_width + margin
        box_y = other_y
    end

    return box_x, box_y
end

function ruler.get_text_box(text, font)
    return font:getWidth(text), font:getHeight()
end

function ruler.draw_grid(cell_width, cell_height, grid_offset_x, grid_offset_y)
    local min_screen_x, min_screen_y = love.graphics.inverseTransformPoint(0, 0)
    local max_screen_x, max_screen_y = love.graphics.inverseTransformPoint(love.graphics.getDimensions())
    local column_count = math.ceil((max_screen_x - min_screen_x) / cell_width + 0.5)
    local row_count = math.ceil((max_screen_y - min_screen_y) / cell_height + 0.5)
    local first_column_x = math.floor(min_screen_x / cell_width + 0.5) * cell_width
    local first_row_y = math.floor(min_screen_y / cell_height + 0.5) * cell_height

    for row = 0, row_count-1 do
        local row_y = first_row_y + cell_height * row
        love.graphics.line(min_screen_x, row_y, max_screen_x, row_y)
    end

    for column = 0, column_count-1 do
        local column_x = first_column_x + cell_width * column
        love.graphics.line(column_x, min_screen_y, column_x, max_screen_y)
    end
end

function ruler.draw_vertical_guide(x)
    local min_screen_x, min_screen_y = love.graphics.inverseTransformPoint(0, 0)
    local max_screen_x, max_screen_y = love.graphics.inverseTransformPoint(love.graphics.getDimensions())

    if x >= min_screen_x and x <= max_screen_x then
        love.graphics.line(x, min_screen_y, x, max_screen_y)
    end
end

function ruler.draw_horizontal_guide(y)
    local min_screen_x, min_screen_y = love.graphics.inverseTransformPoint(0, 0)
    local max_screen_x, max_screen_y = love.graphics.inverseTransformPoint(love.graphics.getDimensions())

    if y >= min_screen_y and y <= max_screen_y then
        love.graphics.line(min_screen_x, y, max_screen_x, y)
    end
end

function ruler.draw_top_down_space(grid_cell_width, grid_cell_height, grid_offset_x, grid_offset_y)
    love.graphics.push()
    love.graphics.translate(grid_offset_x or 0, grid_offset_y or 0)
    love.graphics.setLineStyle('rough')
    love.graphics.setColor(1,1,1,0.25)
    ruler.draw_grid(grid_cell_width, grid_cell_height)
    love.graphics.setColor(1,0.5,0.5,0.35)
    ruler.draw_horizontal_guide(0)
    love.graphics.setColor(0.5,1,0.5,0.35)
    ruler.draw_vertical_guide(0)
    love.graphics.pop()
end

function ruler.grid(column_count, row_count, cell_width, cell_height, grid_offset_x, grid_offset_y)
    local i = 0
    local cell_count = column_count * row_count

    return function()
        local cell_x, cell_y
        local column, row

        if i == cell_count then return end

        column = i % column_count
        row = math.floor(i / column_count)

        cell_x = grid_offset_x + (column) * cell_width
        cell_y = grid_offset_y + (row) * cell_height

        i = i + 1

        return cell_x, cell_y, column+1, row+1
    end
end

return ruler