local animation = require(PATH..'.animation')

local sound_factory = {}

function sound_factory.sine(phase) return math.sin(phase) end
function sound_factory.noise() return math.random() * 2 - 1 end
function sound_factory.square(phase) 
    local sine = math.sin(phase)
    return sine > 0 and 1 or sine < 0 and -1 or 0 
end
function sound_factory.square_antialiased(phase, frequency) 
    local harmonics_count = math.floor(10000 / frequency)
    local signal = 0
    for i = 1, harmonics_count do
        local harmonic_number = (2*(i-1)+1)
        signal = signal + (1/harmonic_number) * math.sin(phase * harmonic_number)
    end
    return signal
end
function sound_factory.triangle(phase)
    phase = math.fmod(phase + 0.5 * math.pi, 2*math.pi)
    return phase <= math.pi and (2 * phase / math.pi - 1) or 2 * (1 - (phase - math.pi) / math.pi) - 1
end
function sound_factory.quarter_duty(phase)
    phase = math.fmod(phase, 2*math.pi)
    return phase < 0.5 * math.pi and 1 or -1
end
function sound_factory.eightth_duty(phase)
    phase = math.fmod(phase, 2*math.pi)
    return phase < 0.25 * math.pi and 1 or -1
end

function sound_factory.new_sound_effect(length, frequency, waveform)
    return {
        waveform = waveform or sound_factory.sine,
        frequency = frequency or 440,
        length = length or 1,
        start_amplitude = 1,
        end_amplitude = 1,
        start_pitch = 1,
        end_pitch = 1,
        amp_easing = animation.ease_quad_out,
        pitch_easing = animation.ease_quad_out,
        pitch_mod_frequency = 1,
        pitch_mod_amplitude = 0,
        pitch_mod_phase = 0,
        amp_mod_frequency = 1,
        amp_mod_amplitude = 0,
        amp_mod_phase = 0,
    }
end

function sound_factory.sweep_up_amplitude(sound_effect, easing)
    sound_effect.start_amplitude = 0
    sound_effect.end_amplitude = 1
    sound_effect.amp_easing = easing or sound_effect.amp_easing
end

function sound_factory.sweep_down_amplitude(sound_effect, easing)
    sound_effect.end_amplitude = 0
    sound_effect.start_amplitude = 1
    sound_effect.amp_easing = easing or sound_effect.amp_easing
end

function sound_factory.modulate_amp(sound_effect, amplitude, frequency, phase)
    sound_effect.amp_mod_amplitude = amplitude
    sound_effect.amp_mod_frequency = frequency
    sound_effect.amp_mod_phase = phase or 0
end

function sound_factory.sweep_up_pitch(sound_effect, end_pitch, easing)
    sound_effect.start_pitch = 1
    sound_effect.end_pitch = end_pitch or 10
    sound_effect.pitch_easing = easing or sound_effect.pitch_easing
end

function sound_factory.sweep_down_pitch(sound_effect, end_pitch, easing)
    sound_effect.end_pitch = end_pitch or 0.1
    sound_effect.start_pitch = 1
    sound_effect.pitch_easing = easing or sound_effect.pitch_easing
end

function sound_factory.modulate_pitch(sound_effect, amplitude, frequency, phase)
    sound_effect.pitch_mod_amplitude = amplitude
    sound_effect.pitch_mod_frequency = frequency
    sound_effect.pitch_mod_phase = phase or 0
end

local function decibel_conversion(level)
    return math.pow(10, (level/20))
end

function sound_factory.render_sound_effect_to(buffer, sound_effect)
    local sample_length = 1 / buffer:getSampleRate()
    local waveform = sound_effect.waveform
    local phase = 0
    volume = volume or 1
    time_offset = time_offset or 0
    sample_offset = math.floor(time_offset * sample_length)

    for i = sample_offset, buffer:getSampleCount() - 1 do
        local time = i * sample_length
        local progress = time / sound_effect.length
        local pitch = (sound_effect.end_pitch - sound_effect.start_pitch) * sound_effect.pitch_easing(progress) + sound_effect.start_pitch
        local pitch_modulation = sound_effect.pitch_mod_amplitude == 0 and 1 or 1 + sound_effect.pitch_mod_amplitude * math.sin(math.pi * 2 * sound_effect.pitch_mod_frequency * time)
        local amplitude = (sound_effect.end_amplitude - sound_effect.start_amplitude) * sound_effect.amp_easing(progress) + sound_effect.start_amplitude
        local amp_modulation = sound_effect.amp_mod_amplitude == 0 and 1 or sound_effect.amp_mod_amplitude * 0.5 * (1 + math.sin(math.pi * 2 * sound_effect.amp_mod_frequency * time))
        local phase_increment = math.pi * 2 * sound_effect.frequency * sample_length * pitch * pitch_modulation
        --local sample = buffer:getSample(i)
        local rendered_signal = waveform(phase, sound_effect.frequency) * amplitude * amp_modulation
        --local mixed_value = mix_with_data and (rendered_signal * volume + sample) or rendered_signal
        buffer:setSample(i, rendered_signal)
        phase = phase + phase_increment
    end
end

function sound_factory.rendered_sound_effect(sound_effect, sample_rate, bit_depth)
    sample_rate = sample_rate or 44100
    bit_depth = bit_depth or 16
    local sample_count = math.ceil(sample_rate * sound_effect.length)
    local buffer = love.sound.newSoundData(sample_count, sample_rate, bit_depth, 1)
    
    sound_factory.render_sound_effect_to(buffer, sound_effect)
    return buffer
end

function sound_factory.fill_buffer(buffer, waveform_name, frequency)
    local waveform = waveforms[waveform_name]
    local sample_length = 1 / buffer:getSampleRate()

    for i = 0, buffer:getSampleCount() - 1 do
        buffer:setSample(i, waveform(i * sample_length, frequency))
    end
end

function sound_factory.render_sound_effect_to_pitchfail(buffer, sound_effect)
    local sample_length = 1 / buffer:getSampleRate()
    local waveform = sound_effect.waveform

    for i = 0, buffer:getSampleCount() - 1 do
        local time = i * sample_length
        local progress = time / sound_effect.length
        local pitch = (sound_effect.end_pitch - sound_effect.start_pitch) * (progress * progress) + sound_effect.start_pitch
        local amplitude = (sound_effect.end_amplitude - sound_effect.start_amplitude) * (progress * progress) + sound_effect.start_amplitude
        buffer:setSample(i, waveform(i * sample_length, sound_effect.frequency * pitch) * amplitude)
    end
end

local SoundPlayer = {}
SoundPlayer.__index = SoundPlayer

function sound_factory.new_player(sound)
    local player = {
        sound = sound,
        sources = {
            love.audio.newSource(sound),
        },
    }
    setmetatable(player, SoundPlayer)
    return player
end

function SoundPlayer:play(pitch, volume)
    pitch = pitch or 1
    volume = volume or 1
    local create_new_source = true
    for i = 1, #self.sources do
        local source = self.sources[i]
        if not source:isPlaying() then
            source:setPitch(pitch)
            source:setVolume(volume)
            source:play()
            create_new_source = false
            break
        end
    end 
    if create_new_source then
        local new_source = love.audio.newSource(self.sound)
        new_source:setPitch(pitch)
        new_source:setVolume(volume)
        new_source:play()
        table.insert(self.sources, new_source)
    end
end

return sound_factory