local ruler = require(PATH..'.ruler')
local collision_shapes = require(PATH..'.collision_shapes')
local pool = require(PATH..'.pool')
local spatial_hash = {}

local _queries_pool = {}
local _region_pool = {}

function spatial_hash.new(bucket_width, bucket_height)
    return {
        _buckets = {},
        _regions = {},
        _bucket_width = bucket_width,
        _bucket_height = bucket_height,
    }
end

local function get_overlapping_buckets(hash, region_x, region_y, region_x2, region_y2)
    local left_bucket, top_bucket, right_bucket, bottom_bucket

    left_bucket, top_bucket = ruler.get_grid_cell_at(region_x, region_y, hash._bucket_width, hash._bucket_height)
    right_bucket, bottom_bucket = ruler.get_grid_cell_at(region_x2, region_y2, hash._bucket_width, hash._bucket_height)

    return left_bucket, top_bucket, right_bucket, bottom_bucket
end

local function create_bucket(hash, bucket_id)
    if hash._buckets[bucket_id] == nil then
        hash._buckets[bucket_id] = {}
    end
end

function spatial_hash.clear(hash)
    for k in pairs(hash._buckets) do hash._buckets[k] = nil end
    for k in pairs(hash._regions) do hash._regions[k] = nil end
end

function spatial_hash.insert_region(hash, region_x, region_y, region_x2, region_y2)
    local region = pool.fetch(_region_pool)
    local region_handle = tostring(region)
    if region.buckets == nil then 
        region.buckets = {} 
    end
    hash._regions[region_handle] = region
    spatial_hash.set_region(hash, region_handle, region_x, region_y, region_x2, region_y2)
    return region_handle
end

function spatial_hash.set_region(hash, region_handle, new_x, new_y, new_x2, new_y2)
    local region = hash._regions[region_handle]
    local left_bucket, top_bucket, right_bucket, bottom_bucket = get_overlapping_buckets(hash, new_x, new_y, new_x2, new_y2)
    
    region.x = new_x
    region.y = new_y
    region.x2 = new_x2
    region.y2 = new_y2

    for i, bucket_id in ipairs(region.buckets) do 
        local bucket = hash._buckets[bucket_id]

        for j, test_handle in ipairs(bucket) do
            if test_handle == region_handle then
                table.remove(bucket, j)
                break
            end
        end
        region.buckets[i] = nil
    end

    for x = left_bucket, right_bucket do
        for y = top_bucket, bottom_bucket do
            local bucket_id = x..','..y
            create_bucket(hash, bucket_id)
            table.insert(hash._buckets[bucket_id], region_handle)
            table.insert(region.buckets, bucket_id)
        end
    end
end

function spatial_hash.set_region_position(hash, region_handle, new_x, new_y)
    local region = hash._regions[region_handle]
    spatial_hash.set_region(hash, region_handle, new_x, new_y, new_x+region.x2-region.x, new_y+region.y2-region.y)
end

function spatial_hash.get_region(hash, region_handle)
    local region = hash._regions[region_handle]
    return region.x, region.y, region.x2, region.y2
end

function spatial_hash.remove_region(hash, region_handle)
    for i, bucket_id in ipairs(region.buckets) do 
        local bucket = hash._buckets[bucket_id]

        for j, test_handle in ipairs(bucket) do
            if test_handle == region_handle then
                table.remove(bucket, j)
                break
            end
        end
        region.buckets[i] = nil
    end
    pool.release(_region_pool, hash._regions[region_handle], true)
    hash._regions[region_handle] = nil
end

function spatial_hash.get_overlapping_regions(hash, region_x, region_y, region_x2, region_y2)
    local left_bucket, top_bucket, right_bucket, bottom_bucket = get_overlapping_buckets(hash, region_x, region_y, region_x2, region_y2)
    local overlapping_regions = pool.fetch(_queries_pool)
    local overlapping_regions_count = 0
    local j = 0

    for x = left_bucket, right_bucket do
        for y = top_bucket, bottom_bucket do
            local bucket_id = x..','..y
            local bucket = hash._buckets[bucket_id]

            if bucket then
                for i = #bucket, 1, -1 do
                    local region_handle = bucket[i]
                    local region = hash._regions[region_handle]

                    if overlapping_regions[region_handle] == nil and collision_shapes.rectangle_rectangle(region.x, region.y, region.x2, region.y2, region_x, region_y, region_x2, region_y2) then
                        table.insert(overlapping_regions, bucket[i])
                        overlapping_regions[region_handle] = true
                    end
                end
            end
        end
    end

    overlapping_regions_count = #overlapping_regions

    return function()
        j = j + 1
        if j <= overlapping_regions_count then return overlapping_regions[j] end
        pool.release(_queries_pool, overlapping_regions)
    end
end

function spatial_hash.get_region_overlapping_regions(hash, region_handle)
    local region = hash._regions[region_handle]
    return spatial_hash.get_overlapping_regions(hash, region.x, region.y, region.x2, region.y2)
end

function spatial_hash.get_neighbouring_regions(hash, region_x, region_y, region_x2, region_y2, max_distance)

end

function spatial_hash.get_region_neighbouring_regions(hash, region_handle, max_distance)
    local region = hash._regions[region_handle]
    spatial_hash.get_neighbouring_regions(hash, region.x, region.y, region.x2, region.y2, max_distance)
end

function spatial_hash.visualize(hash)
    love.graphics.setColor(0.3,1,0,0.7)
    for bucket_id, bucket in pairs(hash._buckets) do
        if #bucket > 0 then
            local bucket_x, bucket_y = string.match(bucket_id, '([%d%-]+),([%d%-]+)')
            love.graphics.rectangle('fill', hash._bucket_width * tonumber(bucket_x), hash._bucket_height * tonumber(bucket_y), hash._bucket_width, hash._bucket_height)
        end
    end

    love.graphics.setColor(1,1,0,0.9)
    for region_handle, region in pairs(hash._regions) do
        love.graphics.rectangle('fill', region.x, region.y, region.x2-region.x, region.y2-region.y)
    end
end

return spatial_hash