local enum = require(PATH..'.enum')
local check = require(PATH..'.check')

local viewport = {
    STRETCH = enum('DISABLED', 'KEEP_RATIO', 'KEEP_WIDTH', 'KEEP_HEIGHT', 'FILL_WINDOW'),
    HORIZONTAL_ALIGNMENT = enum('CENTER', 'LEFT', 'RIGHT'),
    VERTICAL_ALIGNMENT = enum('CENTER', 'TOP', 'BOTTOM'),
}

local _base_width = 640
local _base_height = 480
local _base_ratio = 4 / 3
local _base_ratio_inverse = 3 / 4
local _stretch_mode = viewport.STRETCH.DISABLED
local _horizontal_alignment, _vertical_alignment = viewport.HORIZONTAL_ALIGNMENT.CENTER, viewport.VERTICAL_ALIGNMENT.CENTER
local _cached_viewport_box = { 
    screen_width = -1, screen_height = -1, 
    base_width = -1, base_height = -1, 
    x = 0, y = 0, width = -1, height = -1, 
    halign = -1, valign = -1, 
    transform = love.math.newTransform() 
}

function viewport.set_base_size(width, height)
    check(width, 'width', 'number', 'strictly_positive integer')
    check(height, 'height', 'number', 'strictly_positive integer')
    _base_width = width
    _base_height = height
    _base_ratio = width / height
    _base_ratio_inverse = height / width
end

function viewport.get_base_size()
    return _base_width, _base_height
end

function viewport.set_stretch(stretch_mode)
    viewport.STRETCH('stretch_mode', stretch_mode)
    _stretch_mode = stretch_mode or viewport.STRETCH.DISABLED
end

function viewport.set_alignment(horizontal_alignment, vertical_alignment)
    horizontal_alignment = horizontal_alignment or viewport.HORIZONTAL_ALIGNMENT.CENTER
    vertical_alignment = vertical_alignment or viewport.VERTICAL_ALIGNMENT.CENTER
    viewport.HORIZONTAL_ALIGNMENT('horizontal_alignment', horizontal_alignment)
    viewport.VERTICAL_ALIGNMENT('vertical_alignment', vertical_alignment)
    _horizontal_alignment = horizontal_alignment
    _vertical_alignment = vertical_alignment
end

function viewport.get_box()
    local lgw, lgh = love.graphics.getDimensions()
    local screen_changed = _cached_viewport_box.screen_width ~= lgw or _cached_viewport_box.screen_height ~= lgh
    local viewport_changed = _cached_viewport_box.base_width ~= _base_width or _cached_viewport_box.base_height ~= _base_height
    local alignment_changed = _cached_viewport_box.halign ~= _horizontal_alignment or _cached_viewport_box.valign ~= _vertical_alignment

    if screen_changed or viewport_changed or alignment_changed then
        local window_ratio = lgw / lgh
        local box_x, box_y, box_width, box_height = 0, 0, _base_width, _base_height

        if _stretch_mode == viewport.STRETCH.KEEP_RATIO then
            if window_ratio > _base_ratio then
                box_height = lgh
                box_width = math.floor(lgh * _base_ratio)
            else
                box_width = lgw
                box_height = math.floor(lgw * _base_ratio_inverse)
            end
        elseif _stretch_mode == viewport.STRETCH.FILL_WINDOW then
            box_width = lgw
            box_height = lgh
        end

        if _horizontal_alignment == viewport.HORIZONTAL_ALIGNMENT.RIGHT then
            box_x = lgw - box_width
        elseif _horizontal_alignment == viewport.HORIZONTAL_ALIGNMENT.CENTER then
            box_x = math.floor(0.5 * (lgw - box_width))
        end

        if _vertical_alignment == viewport.VERTICAL_ALIGNMENT.BOTTOM then
            box_y = lgh - box_height
        elseif _vertical_alignment == viewport.VERTICAL_ALIGNMENT.CENTER then
            box_y = math.floor(0.5 * (lgh - box_height))
        end

        _cached_viewport_box.transform:reset()
        _cached_viewport_box.transform:translate(box_x, box_y)
        _cached_viewport_box.transform:scale(box_width / _base_width, box_height / _base_height)

        _cached_viewport_box.screen_width, _cached_viewport_box.screen_height = lgw, lgh
        _cached_viewport_box.halign = _horizontal_alignment
        _cached_viewport_box.valign = _vertical_alignment
        _cached_viewport_box.base_width = _base_width
        _cached_viewport_box.base_height = _base_height
        _cached_viewport_box.x, _cached_viewport_box.y, _cached_viewport_box.width, _cached_viewport_box.height = box_x, box_y, box_width, box_height
    end

    return _cached_viewport_box.x, _cached_viewport_box.y, _cached_viewport_box.width, _cached_viewport_box.height
end

function viewport.get_transform()
    local transform = love.math.newTransform()
    viewport.get_box()
    transform:apply(_cached_viewport_box.transform)
    return transform
end

function viewport.transform(x, y)
    viewport.get_box()
    return _cached_viewport_box.transform:transformPoint(x, y)
end

function viewport.inverse_transform(x, y)
    viewport.get_box()
    return _cached_viewport_box.transform:inverseTransformPoint(x, y)
end

function viewport.apply_transform()
    viewport.get_box()
    love.graphics.applyTransform(_cached_viewport_box.transform)
end

function viewport.push_transform(...)
    love.graphics.push()
    viewport.apply_transform(...)
end

function viewport.get_scale()
    local x, y, w, h = viewport.get_box()
    return math.max(w / _base_width, h / _base_height)
end

function viewport.get_scale_min()
    local x, y, w, h = viewport.get_box()
    return math.min(w / _base_width, h / _base_height)
end

function viewport.get_scale_xy()
    local x, y, w, h = viewport.get_box()
    return w / _base_width, h / _base_height
end

function viewport.get_scale_y()
    local x, y, w, h = viewport.get_box()
    return h / _base_height
end

function viewport.get_scale_x()
    local x, y, w, h = viewport.get_box()
    return w / _base_width
end

function viewport.get_width()
    return _base_width
end

function viewport.get_height()
    return _base_height
end

viewport.set_base_size(love.graphics.getDimensions())

return viewport