Schema = require("libraries/classic"):extend()

function Schema:new()
    self.nodes = {}
    self.cables = {}
    self.values = {}
    self.links = {}
end

function Schema:__tostring()
    return "Schema"
end

function Schema:add_node(node)
    table.insert(self.nodes, node)
end

function Schema:connect(out_node, in_node)
    table.insert(self.cables, Cable(out_node, in_node, self))
    if self:test_loop() then
        self:delete_cable(self.cables[#self.cables])
    else
        self:resolve()
    end
end

function Schema:start_wiring(out_node)
    local cable = Cable(out_node, nil, self)
    table.insert(self.cables, cable)
    cable.is_tail_dragged = true
end

function Schema:update()
    for i = #self.nodes, 1, -1 do
        self.nodes[i]:update()
    end

    for i = #self.cables, 1, -1 do
        self.cables[i]:update()
    end

    if input:pressed("new emitter") then
        self:add_node(EmitterNode(love.mouse.getX(),love.mouse.getY(),self))
    end

    if input:pressed("new screen") then
        self:add_node(ScreenNode(love.mouse.getX(),love.mouse.getY(),self))
    end

    if input:pressed("new refractor") then
        self:add_node(RefractorNode(love.mouse.getX(),love.mouse.getY(),self))
    end

    if input:pressed("new attenuator") then
        self:add_node(AttenuatorNode(love.mouse.getX(),love.mouse.getY(),self))
    end

    if input:pressed("new combiner") then
        self:add_node(CombinerNode(love.mouse.getX(),love.mouse.getY(),self))
    end

    --self:resolve()
end

function Schema:is_overlapping_nodes(node)
    for i = 1, #self.nodes do
        local other = self.nodes[i]

        if other ~= node and node:is_overlapping(other) then
            return true
        end
    end
end

function Schema:adjust_node_position(node)
    local adjust_x, adjust_y = 0, 0

    for i = 1, #self.nodes do
        local other = self.nodes[i]

        if other ~= node then
            local part_adjust_x, part_adjust_y = other:get_overlap(node)

            adjust_x = adjust_x + part_adjust_x
            adjust_y = adjust_y + part_adjust_y
        end
    end

    node.x = node.x + adjust_x
    node.y = node.y + adjust_y
end

function Schema:put_on_top(node)
    for i = #self.nodes, 1, -1 do
        local other = self.nodes[i]

        if node == other then
            table.insert(self.nodes, table.remove(self.nodes, i))
            return
        end
    end
end 

function Schema:is_not_obfuscated(node, x, y)
    for i = #self.nodes, 1, -1 do
        local other = self.nodes[i]

        if node == other then
            return true
        elseif other:is_point_inside(x, y) then
            return false
        end
    end
end

function Schema:rewire_cable(cable)
    local is_successful = false

    for i = 1, #self.nodes do
        local node = self.nodes[i]

        --if node ~= cable.out_node then
            if self:has_available_inputs(node) and node.input_circle:is_overlapping(cable.tail_circle) then
                local is_cloned_link = false

                -- prevents cloned links
                for j = 1, #self.cables do
                    local other = self.cables[j]

                    if other ~= cable then
                        if other.in_node == node and other.out_node == cable.out_node then
                            is_cloned_link = true
                            break
                        end
                    end
                end

                if not is_cloned_link then
                    cable:set_in_node(node)

                    if not self:test_loop() then
                        is_successful = true
                    end
                end

                break
            end
        --end
    end
    
    if not is_successful then
        self:delete_cable(cable)
    else
        self:resolve()
    end
end

local function test_loop_recursive(self, node, visited_nodes, path_stack)
    visited_nodes[node] = true
    table.insert(path_stack, node)
    path_stack[node] = true

    for i = 1, #self.cables do
        local cable = self.cables[i]

        if cable:is_complete() and cable.out_node == node then
            local neighbour = cable.in_node

            if path_stack[neighbour] then
                return true
            end

            if not visited_nodes[neighbour] then
                if test_loop_recursive(self, neighbour, visited_nodes, path_stack) then
                    return true
               end
            end
        end
    end

    table.remove(path_stack)
    path_stack[node] = nil
end 

function Schema:test_loop()
    -- https://www.geeksforgeeks.org/detect-cycle-in-a-graph/
    local visited_nodes, path_stack = {}, {}
    for i = 1, #self.nodes do
        if test_loop_recursive(self, self.nodes[i], visited_nodes, path_stack) then
            return true
        end
    end
    return false
end

function Schema:get_local_color(node)
    if self.values[node] == nil then
        return 0, 0, 0
    else
        return unpack(self.values[node])
    end
end

local function resolve_recursive(self, node, visited_nodes)
    if visited_nodes[node] then return end

    visited_nodes[node] = true
    local input_values = {}

    for i = 1, #self.cables do
        local cable = self.cables[i]

        if cable.in_node == node then
            local neighbour = cable.out_node

            resolve_recursive(self, neighbour, visited_nodes)
        end
    end

    for i = 1, #self.cables do
        local cable = self.cables[i]

        if cable.in_node == node then
            local neighbour = cable.out_node
            local nr, ng, nb = self:get_local_color(neighbour)

            table.insert(input_values, nr)
            table.insert(input_values, ng)
            table.insert(input_values, nb)
        end
    end

    print(node)

    if #input_values >= 3 then
        self.values[node] = {node:get_output_color(unpack(input_values))}
    else
        self.values[node] = {node:get_output_color(0,0,0)}
    end
end 

function Schema:resolve()
    print("RESOLVE")
    local visited_nodes = {}
    for i = 1, #self.nodes do
        resolve_recursive(self, self.nodes[i], visited_nodes)
    end
end

function Schema:delete_cable(cable)
    for i = 1, #self.cables do
        if self.cables[i] == cable then
            table.remove(self.cables, i)
            break
        end
    end
    self:resolve()
end

function Schema:delete_node(node)
    for i = 1, #self.nodes do
        if self.nodes[i] == node then
            table.remove(self.nodes, i)
            break
        end
    end
    for i = #self.cables, 1, -1 do
        if self.cables[i].in_node == node or self.cables[i].out_node == node then
            table.remove(self.cables, i)
        end
    end
    self:resolve()
end

function Schema:get_used_inputs(node)
    local count = 0

    for i = 1, #self.cables do
        if self.cables[i].in_node == node then
            count = count + 1
        end
    end

    return count
end

function Schema:get_used_outputs(node)
    local count = 0

    for i = 1, #self.cables do
        if self.cables[i].out_node == node then
            count = count + 1
        end
    end

    return count
end

function Schema:has_available_inputs(node)
    return node:get_max_inputs() == -1 or self:get_used_inputs(node) < node:get_max_inputs()
end

function Schema:has_available_outputs(node)
    return node:get_max_outputs() == -1 or self:get_used_outputs(node) < node:get_max_outputs()
end

function Schema:draw()
    for i = 1, #self.cables do
        self.cables[i]:draw()
    end

    for i = 1, #self.nodes do
        self.nodes[i]:draw()
    end
end